import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Properties;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageProducer;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.hornetq.api.core.DiscoveryGroupConfiguration;
import org.hornetq.api.core.TransportConfiguration;
import org.hornetq.api.jms.HornetQJMSClient;
import org.hornetq.api.jms.JMSFactoryType;
import org.hornetq.core.remoting.impl.netty.NettyConnectorFactory;
import org.hornetq.core.remoting.impl.netty.TransportConstants;
import org.hornetq.jms.client.HornetQConnectionFactory;

public class QueueExampleForDiscovery {

    public static void main(String[] args) throws Exception {
        new QueueExampleForDiscovery().run();
    }

    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat(
            "dd.MM.yyyy HH:mm:ss;SSS");

    private final ExecutorService executorService = Executors.newCachedThreadPool();

    private Connection connection;

    private final int numberOfMessages = Integer.MAX_VALUE;

    private final String queueName = "foobar";

    private final AtomicBoolean producerIsWorking = new AtomicBoolean(true);

    private final AtomicBoolean consumerIsWorking = new AtomicBoolean(true);

    private enum ConnectionWith {
        JNDI, STATIC, DISCOVERY_GROUP
    }

    private final ConnectionWith connectionWith = ConnectionWith.JNDI;

    private Connection createConnection() throws NamingException, JMSException {
        switch (connectionWith) {
        case DISCOVERY_GROUP:
            return createConnectionWithDiscoveryGroup();
        case JNDI:
            return createConnectionWithJndi();
        case STATIC:
            return createConnectionWithStatic();
        }
        throw new IllegalArgumentException(String.valueOf(connectionWith));
    }

    private Connection createConnectionWithStatic() throws JMSException {
        return createStaticFactory().createConnection();
    }

    private HornetQConnectionFactory createStaticFactory() {

        TransportConfiguration[] transportConfiguration = createTransportConfiguration();

        HornetQConnectionFactory factory = HornetQJMSClient.createConnectionFactoryWithHA(
                JMSFactoryType.CF, transportConfiguration);

        setupFactory(factory);

        return factory;

    }

    private void setupFactory(HornetQConnectionFactory factory) {
        // Justin Bertram says Nope!
        // http://docs.jboss.org/hornetq/2.3.0.Final/docs/user-manual/html/ha.html#ha.automatic.failover
        // factory.setClientFailureCheckPeriod(Duration.ofSeconds(1).toMillis());

        // 39.2.1.1 on
        // http://docs.jboss.org/hornetq/2.3.0.Final/docs/user-manual/html/ha.html#ha.automatic.failover
        factory.setInitialConnectAttempts(5);

        // 34.1
        factory.setConfirmationWindowSize(1000000);

        // 34.1
        factory.setReconnectAttempts(20);
        factory.setRetryIntervalMultiplier(1.5);
        factory.setMaxRetryInterval(8000);
    }

    private TransportConfiguration[] createTransportConfiguration() {
        return new TransportConfiguration[] { createTransportConfiguration("localhost", 62500),
                createTransportConfiguration("localhost", 62510),
        // createTransportConfiguration("localhost", 62520)
        };
    }

    private TransportConfiguration createTransportConfiguration(String host, int port) {
        HashMap<String, Object> map = new HashMap<String, Object>();
        map.put(TransportConstants.HOST_PROP_NAME, host);
        map.put(TransportConstants.PORT_PROP_NAME, String.valueOf(port));
        TransportConfiguration tc = new TransportConfiguration(NettyConnectorFactory.class
                .getName(), map);
        return tc;
    }

    private Connection createConnectionWithJndi() throws NamingException, JMSException {
        // try {
        InitialContext initialContext = getContext("jnp://localhost:62502");
        ConnectionFactory connectionFactory = (ConnectionFactory) initialContext
                .lookup("/ConnectionFactory");
        return connectionFactory.createConnection();
        // } catch (CommunicationException e) {
        // InitialContext initialContext = getContext("jnp://localhost:62512");
        // ConnectionFactory connectionFactory = (ConnectionFactory)
        // initialContext
        // .lookup("/ConnectionFactory");
        // return connectionFactory.createConnection();
        // }
    }

    private InitialContext getContext(String server) throws NamingException {
        Properties props = new Properties();
        props.put("java.naming.factory.initial", "org.jnp.interfaces.NamingContextFactory");
        props.put("java.naming.provider.url", server);
        props.put("java.naming.factory.url.pkgs", "org.jboss.naming:org.jnp.interfaces");
        return new InitialContext(props);
    }

    private Connection createConnectionWithDiscoveryGroup() throws JMSException {
        return createDiscoveryGroupFactory().createConnection();
    }

    private HornetQConnectionFactory createDiscoveryGroupFactory() {
        HornetQConnectionFactory factory = HornetQJMSClient.createConnectionFactoryWithHA(
                getGroupConfiguration(), JMSFactoryType.CF);
        setupFactory(factory);
        return factory;
    }

    private DiscoveryGroupConfiguration getGroupConfiguration() {
        // see
        // http://docs.jboss.org/hornetq/2.2.5.Final/user-manual/en/html/clusters.html#d0e9883
        return new DiscoveryGroupConfiguration("231.7.7.7", 9876);
    }

    public void run() throws InterruptedException, JMSException, NamingException {

        try {
            connection = createConnection();
            connection.start();

            startConsumer();

            startProducer();

            while (producerIsWorking.get() || consumerIsWorking.get()) {
                Thread.sleep(500);
            }

            executorService.shutdown();
            executorService.awaitTermination(10, TimeUnit.SECONDS);
        } finally {
            if (connection != null) {
                connection.stop();
                connection.close();
            }
            executorService.shutdownNow();
        }
    }

    private void startProducer() {

        Callable<Void> task = new Callable<Void>() {

            @Override
            public Void call() {
                try {
                    produce();
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    System.out.println("producer finished");
                    producerIsWorking.set(false);
                }
                return null;
            }

            private void produce() throws JMSException, InterruptedException {
                Session session = connection.createSession(true, Session.AUTO_ACKNOWLEDGE);
                Destination destination = session.createQueue(queueName);
                MessageProducer producer = session.createProducer(destination);

                for (int i = 0; i < numberOfMessages; i++) {
                    try {
                        String text = i + " [" + DATE_FORMAT.format(new Date()) + "]";
                        TextMessage message = session.createTextMessage(text);
                        System.out.println("sending message " + text);
                        producer.send(message);
                        session.commit();
                    } catch (JMSException e) {
                        System.out.println("during send " + e);
                        Thread.sleep(10_000);
                        session.rollback();
                    }
                    Thread.sleep(20);
                }

                producer.close();
                session.close();

            }

        };

        executorService.submit(task);

    }

    private void startConsumer() {

        Callable<Void> task = new Callable<Void>() {

            @Override
            public Void call() {
                try {
                    consume();
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    System.out.println("consumer finished");
                    consumerIsWorking.set(false);
                }
                return null;
            }

            private void consume() throws JMSException, InterruptedException {
                Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                Destination destination = session.createQueue(queueName);
                javax.jms.MessageConsumer consumer = session.createConsumer(destination);

                while (true) {
                    try {
                        Message message = consumer.receive(5000L);
                        if (message == null) {
                            if (producerIsWorking.get()) {
                                System.out.println("received nothing, wait and go on receiving");
                                Thread.sleep(5000);
                                continue;
                            } else {
                                System.out.println("received nothing, stop");
                                break;
                            }
                        }
                        if (message instanceof TextMessage) {
                            System.out.println("recieved: " + ((TextMessage) message).getText());
                        } else if (message instanceof ObjectMessage) {
                            System.out
                                    .println("recieved: " + ((ObjectMessage) message).getObject());
                        } else {
                            System.out.println("recieved: " + message);
                        }
                    } catch (JMSException e) {
                        System.out.println("during receive " + e);
                        if (!producerIsWorking.get()) {
                            break;
                        }
                        Thread.sleep(10_000);
                        continue;
                    }
                }

                consumer.close();
                session.close();

            }

        };

        executorService.submit(task);

    }
}
